#include "service.grpc.pb.h"

#include <boost/asio.hpp>
#include <boost/thread.hpp>
#include <grpcpp/grpcpp.h>

using namespace std;

boost::asio::io_service io_service;
boost::asio::io_service::work work(io_service);
boost::thread_group threadpool;

shared_ptr<grpc::Channel> channel;
unique_ptr<TestService::Stub> stub;
shared_ptr<grpc::ClientContext> context;
shared_ptr<grpc::ClientReaderWriter<Data, Data>> stream;

const auto MAX_BACKLOG = 50000;
atomic_ulong backlog;

ulong num_read = 0;
ulong num_written = 0;

void
grpc_write_thread()
{
    string bytes(200, 'X');
    Data data;
    data.set_payload(bytes);
    while (true) {
        if (num_written % 10000 == 0) {
            //cerr << "wrote " << num_written << endl;
            //cerr << "backlog: " << backlog << endl;
        }
        if (backlog < MAX_BACKLOG) {
            if (!stream->Write(data)) {
                cerr << "client write error" << endl;
                _exit(1);
            }
            ++backlog;
            ++num_written;
        }
    }
}

void
grpc_read_thread()
{
    while (true) {
        if (backlog > 0) {
            // Read raw buffer and initialize frame header.
            Data data;
            if (!stream->Read(&data)) {
                cerr << "client read error" << endl;
                _exit(1);
            }
            --backlog;
            ++num_read;
            if (num_read % 10000 == 0) {
                //cerr << "read " << num_read << endl;
            }
        }
    }
}

int
main(int argc, char* argv[])
{
    grpc::ChannelArguments cargs;
    cargs.SetCompressionAlgorithm(GRPC_COMPRESS_NONE);
    cargs.SetInt(GRPC_ARG_INITIAL_RECONNECT_BACKOFF_MS, 100);
    cargs.SetInt(GRPC_ARG_MIN_RECONNECT_BACKOFF_MS, 100);
    cargs.SetInt(GRPC_ARG_MAX_RECONNECT_BACKOFF_MS, 500);

    channel = grpc::CreateCustomChannel("localhost:7777", grpc::InsecureChannelCredentials(), cargs);
    stub = TestService::NewStub(channel);
    context = make_shared<grpc::ClientContext>();
    stream = stub->DataStream(context.get());

    backlog = 0;

    threadpool.create_thread([]{ io_service.run(); });
    threadpool.create_thread([]{ io_service.run(); });
    io_service.post([]{ grpc_write_thread(); });
    io_service.post([]{ grpc_read_thread(); });

    threadpool.join_all();
}
